%%%----------------------------------------------------------------------
%%% File    : ethbr_center.erl
%%% Author  : Scott Lystig Fritchie <slfritchie@snookles.com>
%%% Purpose : Ethernet bridge central process
%%% Created : 01 Aug 2002 by Scott Lystig Fritchie <slfritchie@snookles.com>
%%%----------------------------------------------------------------------

-module(ethbr_center).
-compile([verbose, report_errors, report_warnings, trace, debug]).

-behaviour(gen_server).

-define(NAME, ?MODULE).
-define(Timeout, infinity).

%% External exports
-export([start_link/1]).
-export([example/2]).

-include("ethbridge.hrl").

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2,
	 code_change/3]).

-record(state, {
	  iface_sup,				% atom: sup's registered name
	  iface_sup_up = false,			% true|false
	  ifacelist = [],			% List of iface_state records
	  timerref,				% Periodic timer reference
	  mactab				% Mac table
	 }).

%%%----------------------------------------------------------------------
%%% API
%%%----------------------------------------------------------------------
start_link(ArgList) ->
    io:format("XXX ~w:start_link ArgList = ~w\n", [?MODULE, ArgList]),
    case lists:keysearch(center_name, 1, ArgList) of
	{value, {_, CName}} ->
	    gen_server:start_link({local, CName}, ?MODULE, ArgList, []);
	_ ->
	    {error, {center_name, not_found}}
    end.

example(A1, A2) ->
    Timeout = ?Timeout,
    gen_server:call(?NAME, {example, A1, A2}, Timeout).

%%%----------------------------------------------------------------------
%%% Callback functions from gen_server
%%%----------------------------------------------------------------------

%%----------------------------------------------------------------------
%% Func: init/1
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          ignore               |
%%          {stop, Reason}
%%----------------------------------------------------------------------
init(StartArgs) ->
    process_flag(trap_exit, true),
    io:format("XXX ~w:init StartArgs = ~w\n", [?MODULE, StartArgs]),
    {value, {_, ISup}} = lists:keysearch(iface_sup_name, 1, StartArgs),
    IList = case lists:keysearch(interfaces_from, 1, StartArgs) of
		{value, {_, env}} ->
		    {ok, EnvIList} = application:get_env(ethbridge,
							 interfaces),
		    io:format("XXX EnvIList = ~p\n", [EnvIList]),
		    make_iface_state_list(EnvIList);
		_ ->
		    []
	    end,
    {ok, TimerRef} = timer:send_interval(5000, {check_interfaces}),
    MacTab = new_mactab(),
    {ok, #state{iface_sup = ISup, ifacelist = IList, mactab = MacTab}}.

%%----------------------------------------------------------------------
%% Func: handle_call/3
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_call(Request, From, State) ->
    Reply = ok,
    {reply, Reply, State}.

%%----------------------------------------------------------------------
%% Func: handle_cast/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_cast(ethbr_iface_sup_started, State) ->
    io:format("XXXYYYXXX ~w: ~s:handle_cast iface sup started msg\n", [self(), ?MODULE]),
    NewState = add_local_ifaces(State),
    {noreply, NewState};
handle_cast({rx_pkt, FromPid, Pkt}, State) ->
    io:format("XXXYYYXXX ~w: ~s:handle_cast got from ~w pkt ~w\n", [self(), ?MODULE, FromPid, Pkt]),
    NewState = do_bridge(FromPid, Pkt, State),
    {noreply, NewState};
handle_cast(Msg, State) ->
    io:format("XXXYYYXXX ~w: ~s:handle_cast got ~w\n", [self(), ?MODULE, Msg]),
    {noreply, State}.

%%----------------------------------------------------------------------
%% Func: handle_info/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_info({check_interfaces}, State) ->
    io:format("XXXYYYXXX ~w: ~s:handle_info got check_interfaces\n", [self(), ?MODULE]),
    NewState = check_interfaces(State),
    {noreply, NewState};
handle_info({'EXIT', Pid, Reason}, State) ->
    io:format("XXXYYYXXX ~w: ~s:handle_info ~w died reason ~w\n", [self(), ?MODULE, Pid, Reason]),
    NewState = clear_dead_pids(Pid, State),
    NewState2 = recalc_forwarding(NewState),
    {noreply, NewState2};
handle_info(Info, State) ->
    io:format("XXXYYYXXX ~w: ~s:handle_info got ~w\n", [self(), ?MODULE, Info]),
    {noreply, State}.

%%----------------------------------------------------------------------
%% Func: terminate/2
%% Purpose: Shutdown the server
%% Returns: any (ignored by gen_server)
%%----------------------------------------------------------------------
terminate(Reason, State) ->
    timer:cancel(State#state.timerref),
    ok.

%%----------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Change code version
%% Returns: {ok, NewState}
%%----------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
    {ok, State}.

%%%----------------------------------------------------------------------
%%% Internal functions
%%%----------------------------------------------------------------------

make_iface_state_list(IL) ->
    make_iface_state_list(IL, []).
make_iface_state_list([], Acc) ->
    io:format("XXX ~w:make_iface_state_list: Acc = ~p\n", [?MODULE, Acc]),
    Acc;
make_iface_state_list([I|Is], Acc) ->
    {interface, IName, BPFProg, Options} = I,
    IF = #iface{name = IName, bpf_prog = BPFProg,
		options = Options, local = true},
    make_iface_state_list(Is, [IF|Acc]).


do_bridge(FromPid, Pkt, State) ->
    case forward_calc(FromPid, Pkt, State) of
	%% QQQ We're assuming that the same packet gets sent to
	%% all interfaces: no per-interface packet editing.
	{ok, NewPkt, ToList, NewState} ->
	    io:format("QQQ AAA do_bridge: ToList = ~w\n", [ToList]),
	    %% QQQ NewPktSize = size(NewPkt),
	    %%lists:map(fun(To) ->
		%%	      %% QQQ Use pattern match on {ok, NewPktSize} ??
		%%	      libnet_drv:libnet_write_link_layer(To#iface.tx_pid,
		%%					     To#iface.tx_link,
		%%					     To#iface.name,
		%%					     NewPkt) end,
		%%      ToList),
	    NewState2 = forward_packet(NewPkt, ToList, NewState),
	    NewState2;
	{nope, NewState} ->
	    NewState
    end.

forward_calc(FromPid, Pkt, State) ->
    {SrcMAC, DstMAC} = eth_srcdstmac(Pkt),
    {ToList, NewState} =
	case DstMAC of
	    <<16#ff:8, 16#ff:8, 16#ff:8, 16#ff:8, 16#ff:8,
	    16#ff:8>> ->
		%% Broadcast
		all_ifaces_except(FromPid, State);
	    <<A:8, _/binary>> when A band 1 == 1 ->
		%% Multicast
		all_ifaces_except(FromPid, State);
	    _ ->
		case lookup_mac_insert(SrcMAC, FromPid, DstMAC, State) of
		    {ok, FromPid} ->
			io:format("QQQ forward_calc: DstMAC on same side\n"),
			{[], State};		% DstMAC on same side
		    {ok, Pid} ->
			{value, IFace} = lists:keysearch(Pid, #iface.rx_pid,
							 State#state.ifacelist),
			io:format("QQQ forward_calc: send to ~w\n", [IFace]),
			{[IFace], State};
		    _ ->
			io:format("QQQ forward_calc: not found, send to all\n"),
			all_ifaces_except(FromPid, State)
		end
	end,
    {ok, Pkt, ToList, NewState}.

%%%
%%% MAC address table functions.  We'll use an ETS table but use a
%%% full state-pass-through style anyway: who knows if things will
%%% change someday?
%%%
%%% mactab key = source MAC address
%%% mactab value = PID of interface that MAC was heard from.

new_mactab() ->
    ets:new(mactab, [set, private]).

clear_mactab(Pid, State) ->
    ets:match_delete(State#state.mactab, {'_', Pid}),
    State.

lookup_mac_insert(SrcMAC, SrcPid, DstMAC, State) ->
    case ets:lookup(State#state.mactab, DstMAC) of
	[{DstMAC, Pid}] ->
	    {ok, Pid};
	[] ->
	    ets:insert(State#state.mactab, {SrcMAC, SrcPid}),
	    not_found
    end.

eth_srcdstmac(<<D:6/binary, S:6/binary, _/binary>>) ->
    {S, D}.

%%%
%%% Local interface stuff
%%%

%%    NewState = add_local_ifaces(State),

add_local_ifaces(State) ->
    add_local_ifaces(State#state.ifacelist, [], State).
add_local_ifaces([], Acc, State) ->
    io:format("XXX ~w:add_local_ifaces: final Acc = ~p\n", [?MODULE, Acc]),
    State#state{ifacelist = Acc};
add_local_ifaces([I|Is], Acc, State) when I#iface.rx_pid =/= undefined,
					  I#iface.tx_pid =/= undefined ->
    add_local_ifaces(Is, [I|Acc], State);
add_local_ifaces([I|Is], Acc, State) ->
    case ethbr_pcap:start({self(), I}) of
	{ok, RxPid} ->
	    io:format("XXX ~w:add_local_ifaces: RxPid = ~p\n", [?MODULE, RxPid]),
	    IName = I#iface.name,
	    F = fun () ->
			{ok, P} = libnet_drv:start(),
			{ok, LinkInt} = libnet_drv:libnet_open_link_interface(P,IName),
			{ok, P, LinkInt}
		end,
	    case catch F() of
		{ok, OutPort, OutLink} ->
		    add_local_ifaces(Is, [I#iface{rx_pid = RxPid, tx_pid = OutPort,
						  tx_link = OutLink}|Acc], State);
		Err ->
		    exit(RxPid, shutdown),
		    io:format("XXX ~w:add_local_ifaces: Err = ~p\n", [?MODULE, Err]),
		    add_local_ifaces(Is, [I|Acc], State)
	    end;
	Err ->
	    io:format("XXX ~w:add_local_ifaces: Err 2 = ~p\n", [?MODULE, Err]),
	    add_local_ifaces(Is, [I|Acc], State)
    end.

check_interfaces(State) ->
    %% add_local_ifaces/1 is smart enough to ignore interfaces that are
    %% already up.
    add_local_ifaces(State).

clear_dead_pids(Pid, State) ->
    clear_dead_pids(Pid, State#state.ifacelist, [], State).
clear_dead_pids(Pid, [], Acc, State) ->
    State#state{ifacelist = Acc};
%% QQQ Am I just being daft, or shouldn't this be legal?
%% clear_dead_pids(Pid, [I|Is], Acc, State) when I#iface.tx_pid == Pid orelse
%%                                               I#iface.rx_pid == Pid -> 
clear_dead_pids(Pid, [I|Is], Acc, State) when I#iface.tx_pid == Pid;
					      I#iface.rx_pid == Pid ->
    io:format("XXX ~w:clear_dead_pids: found Pid ~w in ~p\n", [?MODULE, Pid, I]),
    NMacTab = case I#iface.rx_pid of
		  Pid ->
		      clear_mactab(Pid, State#state.mactab);
		  _ ->
		      State#state.mactab
	      end,
    shutdown_other_half_of_interface(Pid, I#iface.rx_pid, I#iface.local, I),
    clear_dead_pids(Pid, Is, [I#iface{rx_pid = undefined, tx_pid = undefined,
				      tx_link = undefined}|Acc],
		    State#state{mactab = NMacTab});
clear_dead_pids(Pid, [I|Is], Acc, State) ->
    clear_dead_pids(Pid, Is, [I|Acc], State).

recalc_forwarding(State) ->
    State.

% Args: Deceased pid, rx pid, is local interface, interface record

shutdown_other_half_of_interface(DeadPid, DeadPid, true, I) ->
    libnet_drv:libnet_close_link_interface(I#iface.tx_pid, I#iface.tx_link),
    libnet_drv:shutdown(I#iface.tx_pid);
shutdown_other_half_of_interface(DeadPid, DeadPid, false, I) ->
    exit(I#iface.rx_pid, shutdown);
shutdown_other_half_of_interface(DeadPid, RxPid, _, I) ->
    exit(I#iface.tx_pid, shutdown).

all_ifaces_except(Pid, State) ->
    {all_ifaces_except(Pid, State#state.ifacelist, []), State}.
all_ifaces_except(Pid, [], Acc) ->
    Acc;
all_ifaces_except(Pid, [I|Is], Acc) ->
    case I#iface.rx_pid of
	Pid ->
	    all_ifaces_except(Pid, Is, Acc);
	_ ->
	    all_ifaces_except(Pid, Is, [I|Acc])
    end.
	    
forward_packet(Pkt, [], State) ->
    State;
forward_packet(Pkt, [I|Is], State) ->
    PktSize = size(Pkt),
    {ok, PktSize} = libnet_drv:libnet_write_link_layer(I#iface.tx_pid,
						       I#iface.tx_link,
						       I#iface.name,
						       Pkt),
    forward_packet(Pkt, Is, State).

