%%%----------------------------------------------------------------------
%%% File    : ethbr_pcap.erl
%%% Author  : Scott Lystig Fritchie <slfritchie@snookles.com>
%%% Purpose : Receive/libpcap side of interface
%%% Created : 02 Aug 2002 by Scott Lystig Fritchie <slfritchie@snookles.com>
%%%
%%% QQQ This module isn't very OTP-friendly.
%%%----------------------------------------------------------------------

-module(ethbr_pcap).
-compile([verbose, report_errors, report_warnings, trace, debug_info]).

-behaviour(gen_server).

-include("ethbridge.hrl").
-include("libpcap_drv.hrl").

-define(NAME, ?MODULE).
-define(Timeout, infinity).
-define(MY_OS, "FreeBSD").
%-define(MY_OS, "Linux").

%% External exports
-export([start/1]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2,
	 code_change/3]).

-record(state, {
	  port,
	  pcap_iface,
	  center
	 }).

%%%----------------------------------------------------------------------
%%% API
%%%----------------------------------------------------------------------
start(Tuple) ->
    gen_server:start_link(?MODULE, Tuple, []).

%%%----------------------------------------------------------------------
%%% Callback functions from gen_server
%%%----------------------------------------------------------------------

%%----------------------------------------------------------------------
%% Func: init/1
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          ignore               |
%%          {stop, Reason}
%%----------------------------------------------------------------------
init({CenterPid, IFace}) ->
    io:format("XXX ~w:init: Center = ~w, IFace = ~p\n", [?MODULE, CenterPid, IFace]),
    case start_pcap(IFace) of
	{ok, Port, PcapIFace} ->
	    self() ! enter_loop,
	    {ok, #state{port = Port, pcap_iface = PcapIFace,
		        center = CenterPid}};
	Err ->
	    {error, {start_pcap_said, Err}}
    end.

%%----------------------------------------------------------------------
%% Func: handle_call/3
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_call(Request, From, State) ->
    Reply = ok,
    {reply, Reply, State}.

%%----------------------------------------------------------------------
%% Func: handle_cast/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_cast(Msg, State) ->
    io:format("XXXYYYXXX ~w: ~s:handle_cast got ~w\n", [self(), ?MODULE, Msg]),
    {noreply, State}.

%%----------------------------------------------------------------------
%% Func: handle_info/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_info(enter_loop, State) ->
    io:format("XXXYYYXXX ZZZ ~w: ~s:handle_info got enter_loop\n", [self(), ?MODULE]),
    do_loop(State#state.port, State#state.pcap_iface, State#state.center);
handle_info(Info, State) ->
    io:format("XXXYYYXXX ~w: ~s:handle_info got ~w\n", [self(), ?MODULE, Info]),
    {noreply, State}.

%%----------------------------------------------------------------------
%% Func: terminate/2
%% Purpose: Shutdown the server
%% Returns: any (ignored by gen_server)
%%----------------------------------------------------------------------
terminate(Reason, State) ->
    ok.

%%----------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Change code version
%% Returns: {ok, NewState}
%%----------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
    {ok, State}.

%%%----------------------------------------------------------------------
%%% Internal functions
%%%----------------------------------------------------------------------

start_pcap(IFace) when record(IFace, iface) ->
    IName = IFace#iface.name,
    {ok, Port} = libpcap_drv:start(),
    {ok, Int} = libpcap_drv:pcap_open_live(Port, IName,
					   1514, 1, 15000), %QQQ too big!
    case IFace#iface.bpf_prog of
	"" ->
	    ok;
	Prog ->
	    {ok, {_Addr, Mask}} = libpcap_drv:pcap_lookupnet(Port, IName),
	    XXX = libpcap_drv:my_pcap_compile(Port, Int,
						       IFace#iface.bpf_prog,
						       1, Mask),
	    io:format("XXX QQQ YYY ZZZ XXX = ~w\n", [XXX]),
	    {ok, BPFBin} = XXX,
	    {ok, 0} = libpcap_drv:pcap_setfilter(Port, Int, BPFBin),
	    ok
    end,
    case ?MY_OS of
        "FreeBSD" ->
            {ok, 0} = libpcap_drv:my_set_see_sent(Port, Int, 0),
            {ok, 0} = libpcap_drv:my_set_immediate(Port, Int, 1);
        _ ->
            ok
    end,
    %%XXX {ok, 0} = libpcap_drv:debug(Port, 0),
    {ok, Port, Int}.

do_loop(Port, PcapIFace, CenterPid) ->
    case libpcap_drv:pcap_next(Port, PcapIFace) of
	{ok, {{_Pkt_s, _Pkt_us}, Pkt}} ->
	    io:format("XXX ~w:do_loop: QQQ YYY ZZZ XXX = ~w\n", [?MODULE, Pkt]),
	    gen_server:cast(CenterPid, {rx_pkt, self(), Pkt});
	{error, {}} ->
	    ok;
	Err ->
	    io:format("XXX ~s:do_loop got ~w\n", [?MODULE, Err])
    end,
    do_loop(Port, PcapIFace, CenterPid).



	       
